@@ -16,6 +16,8 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#include <stdarg.h>
+
 #include <glib/gi18n.h>
 
 #include "gom-filter.h"
@@ -34,8 +36,7 @@ struct _GomFilterPrivate
    GParamSpec *pspec;
    GType type;
 
-   GomFilter *left;
-   GomFilter *right;
+   GQueue *subfilters;
 };
 
 enum
@@ -58,7 +59,8 @@ static const gchar *gOperators[] = {
    ">=",
    "<",
    "<=",
-   "LIKE"
+   "LIKE",
+   "GLOB"
 };
 
 /**
@@ -117,6 +119,55 @@ gom_filter_new_for_param (GType          resource_type,
    return filter;
 }
 
+static GomFilter *
+gom_filter_new_for_subfilters_full (GomFilterMode  mode,
+                                    GomFilter     *first,
+                                    va_list        filters)
+{
+   GomFilter *filter, *f;
+
+   g_return_val_if_fail(GOM_IS_FILTER(first), NULL);
+
+   filter = g_object_new(GOM_TYPE_FILTER, "mode", mode, NULL);
+   filter->priv->subfilters = g_queue_new();
+   g_queue_push_tail(filter->priv->subfilters, g_object_ref(first));
+
+   f = va_arg(filters, GomFilter*);
+
+   while (f != NULL) {
+      g_return_val_if_fail(GOM_IS_FILTER(f), NULL);
+      g_queue_push_tail(filter->priv->subfilters, g_object_ref(f));
+
+      f = va_arg(filters, GomFilter*);
+   }
+
+   return filter;
+}
+
+static GomFilter *
+gom_filter_new_for_subfilters_fullv (GomFilterMode   mode,
+                                     GomFilter     **filter_array)
+{
+   GomFilter *filter, *f;
+   guint i;
+
+   filter = g_object_new(GOM_TYPE_FILTER, "mode", mode, NULL);
+   filter->priv->subfilters = g_queue_new();
+
+   i = 0;
+   f = filter_array[i];
+
+   while (f != NULL) {
+      g_return_val_if_fail(GOM_IS_FILTER(f), NULL);
+      g_queue_push_tail(filter->priv->subfilters, g_object_ref(f));
+
+      i += 1;
+      f = filter_array[i];
+   }
+
+   return filter;
+}
+
 GomFilter *
 gom_filter_new_like (GType         resource_type,
                      const gchar  *property_name,
@@ -125,6 +176,14 @@ gom_filter_new_like (GType         resource_type,
   return gom_filter_new_for_param(resource_type, property_name, GOM_FILTER_LIKE, value);
 }
 
+GomFilter *
+gom_filter_new_glob (GType         resource_type,
+                     const gchar  *property_name,
+                     const GValue *value)
+{
+  return gom_filter_new_for_param(resource_type, property_name, GOM_FILTER_GLOB, value);
+}
+
 GomFilter *
 gom_filter_new_eq (GType         resource_type,
                    const gchar  *property_name,
@@ -173,7 +232,7 @@ gom_filter_new_lte (GType         resource_type,
    return gom_filter_new_for_param(resource_type, property_name, GOM_FILTER_LTE, value);
 }
 
-GomFilterMode
+static GomFilterMode
 gom_filter_get_mode (GomFilter *filter)
 {
    g_return_val_if_fail(GOM_IS_FILTER(filter), 0);
@@ -225,21 +284,51 @@ get_table (GParamSpec *pspec,
 GomFilter *
 gom_filter_new_and (GomFilter *left,
                     GomFilter *right)
+{
+   GomFilter *filter_array[3] = { left, right, NULL };
+
+   return gom_filter_new_for_subfilters_fullv(GOM_FILTER_AND, filter_array);
+}
+
+/**
+ * gom_filter_new_and_full: (constructor)
+ * @first: (in): A #GomFilter.
+ * @...: (in): A %NULL-terminated list of #GomFilter.
+ *
+ * Creates a new filter that requires that all filters passed as arguments
+ * equate to #TRUE.
+ *
+ * Returns: (transfer full): A #GomFilter.
+ */
+GomFilter *
+gom_filter_new_and_full (GomFilter *first,
+                         ...)
 {
    GomFilter *filter;
+   va_list args;
 
-   g_return_val_if_fail(GOM_IS_FILTER(left), NULL);
-   g_return_val_if_fail(GOM_IS_FILTER(right), NULL);
-
-   filter = g_object_new(GOM_TYPE_FILTER,
-                         "mode", GOM_FILTER_AND,
-                         NULL);
-   filter->priv->left = g_object_ref(left);
-   filter->priv->right = g_object_ref(right);
+   va_start(args, first);
+   filter = gom_filter_new_for_subfilters_full(GOM_FILTER_AND, first, args);
+   va_end(args);
 
    return filter;
 }
 
+/**
+ * gom_filter_new_and_fullv: (constructor)
+ * @filter_array: (in): A %NULL-terminated array of #GomFilter.
+ *
+ * Creates a new filter that requires that all filters passed as arguments
+ * equate to #TRUE.
+ *
+ * Returns: (transfer full): A #GomFilter.
+ */
+GomFilter *
+gom_filter_new_and_fullv (GomFilter **filter_array)
+{
+   return gom_filter_new_for_subfilters_fullv(GOM_FILTER_AND, filter_array);
+}
+
 /**
  * gom_filter_new_or: (constructor)
  * @left: (in): A #GomFilter.
@@ -253,30 +342,61 @@ gom_filter_new_and (GomFilter *left,
 GomFilter *
 gom_filter_new_or (GomFilter *left,
                    GomFilter *right)
+{
+   GomFilter *filter_array[3] = { left, right, NULL };
+
+   return gom_filter_new_for_subfilters_fullv(GOM_FILTER_OR, filter_array);
+}
+
+/**
+ * gom_filter_new_or_full: (constructor)
+ * @first: (in): A #GomFilter.
+ * @...: (in): A %NULL-terminated list of #GomFilter.
+ *
+ * Creates a new filter that requires either of the filters passed as
+ * arguments equate to #TRUE.
+ *
+ * Returns: (transfer full): A #GomFilter.
+ */
+GomFilter *
+gom_filter_new_or_full (GomFilter *first,
+                        ...)
 {
    GomFilter *filter;
+   va_list args;
 
-   g_return_val_if_fail(GOM_IS_FILTER(left), NULL);
-   g_return_val_if_fail(GOM_IS_FILTER(right), NULL);
-
-   filter = g_object_new(GOM_TYPE_FILTER,
-                         "mode", GOM_FILTER_OR,
-                         NULL);
-   filter->priv->left = g_object_ref(left);
-   filter->priv->right = g_object_ref(right);
+   va_start(args, first);
+   filter = gom_filter_new_for_subfilters_full(GOM_FILTER_OR, first, args);
+   va_end(args);
 
    return filter;
 }
 
+/**
+ * gom_filter_new_or_fullv: (constructor)
+ * @filter_array: (in): A %NULL-terminated array of #GomFilter.
+ *
+ * Creates a new filter that requires either of the filters passed as
+ * arguments equate to #TRUE.
+ *
+ * Returns: (transfer full): A #GomFilter.
+ */
+GomFilter *
+gom_filter_new_or_fullv (GomFilter **filter_array)
+{
+   return gom_filter_new_for_subfilters_fullv(GOM_FILTER_OR, filter_array);
+}
+
 gchar *
 gom_filter_get_sql (GomFilter  *filter,
                     GHashTable *table_map)
 {
    GomFilterPrivate *priv;
    gchar *table;
-   gchar *left;
-   gchar *right;
-   gchar *ret;
+   GomFilter *f;
+   gchar **sqls;
+   gint i, len;
+   gchar *sep, *s, *ret;
 
    g_return_val_if_fail(GOM_IS_FILTER(filter), NULL);
 
@@ -292,6 +412,7 @@ gom_filter_get_sql (GomFilter  *filter,
    case GOM_FILTER_LT:
    case GOM_FILTER_LTE:
    case GOM_FILTER_LIKE:
+   case GOM_FILTER_GLOB:
       table = get_table(priv->pspec, priv->type, table_map);
       ret = g_strdup_printf("'%s'.'%s' %s ?",
                             table,
@@ -301,11 +422,29 @@ gom_filter_get_sql (GomFilter  *filter,
       return ret;
    case GOM_FILTER_AND:
    case GOM_FILTER_OR:
-      left = gom_filter_get_sql(priv->left, table_map);
-      right = gom_filter_get_sql(priv->right, table_map);
-      ret = g_strdup_printf("%s %s %s", left, gOperators[priv->mode], right);
-      g_free(left);
-      g_free(right);
+      len = g_queue_get_length(priv->subfilters);
+      sqls = g_new(gchar *, 1 + len);
+
+      for (i = 0; i < len; i++) {
+         f = g_queue_peek_nth(priv->subfilters, i);
+         s = gom_filter_get_sql(f, table_map);
+
+         if ((f->priv->mode == GOM_FILTER_AND) || (f->priv->mode == GOM_FILTER_OR)) {
+            gchar *tmp = g_strdup_printf("(%s)", s);
+            g_free(s);
+            s = tmp;
+         }
+
+         sqls[i] = s;
+      }
+      sqls[i] = NULL;
+
+      sep = g_strdup_printf(" %s ", gOperators[priv->mode]);
+      ret = g_strjoinv(sep, sqls);
+
+      g_free(sep);
+      g_strfreev(sqls);
+
       return ret;
    default:
       break;
@@ -349,8 +488,10 @@ GArray *
 gom_filter_get_values (GomFilter *filter)
 {
    GomFilterPrivate *priv;
+   GomFilter *f;
    GArray *tmp;
    GArray *va;
+   gint i, len;
 
    g_return_val_if_fail(GOM_IS_FILTER(filter), NULL);
 
@@ -370,7 +511,8 @@ gom_filter_get_values (GomFilter *filter)
    case GOM_FILTER_GTE:
    case GOM_FILTER_LT:
    case GOM_FILTER_LTE:
-   case GOM_FILTER_LIKE: {
+   case GOM_FILTER_LIKE:
+   case GOM_FILTER_GLOB: {
       GValue v = { 0 };
       g_value_init(&v, G_VALUE_TYPE(&priv->value));
       g_value_copy(&priv->value, &v);
@@ -381,9 +523,14 @@ gom_filter_get_values (GomFilter *filter)
    }
    case GOM_FILTER_AND:
    case GOM_FILTER_OR:
-      va = gom_filter_get_values(priv->left);
-      tmp = gom_filter_get_values(priv->right);
-      join_value_array(va, tmp);
+      len = g_queue_get_length(priv->subfilters);
+      va = g_array_new(FALSE, FALSE, sizeof(GValue));
+
+      for (i = 0; i < len; i++) {
+         f = g_queue_peek_nth(priv->subfilters, i);
+         tmp = gom_filter_get_values(f);
+         join_value_array(va, tmp);
+      }
 
       return va;
    default:
@@ -416,6 +563,9 @@ gom_filter_finalize (GObject *object)
       g_value_unset(&priv->value);
    }
 
+   if (priv->subfilters != NULL)
+      g_queue_free_full(priv->subfilters, g_object_unref);
+
    G_OBJECT_CLASS(gom_filter_parent_class)->finalize(object);
 }
 
@@ -504,7 +654,7 @@ gom_filter_class_init (GomFilterClass *klass)
    gParamSpecs[PROP_SQL] =
       g_param_spec_string("sql",
                           _("SQL"),
-                          _("The sql for the filter."),
+                          _("The SQL for the filter."),
                           NULL,
                           G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY);
    g_object_class_install_property(object_class, PROP_SQL,
@@ -543,6 +693,7 @@ gom_filter_mode_get_type (void)
       { GOM_FILTER_LT,  "GOM_FILTER_LT",  "LT" },
       { GOM_FILTER_LTE, "GOM_FILTER_LTE", "LTE" },
       { GOM_FILTER_LIKE, "GOM_FILTER_LIKE", "LIKE" },
+      { GOM_FILTER_GLOB, "GOM_FILTER_GLOB", "GLOB" },
       { 0 }
    };
 
