@@ -26,6 +26,11 @@ G_BEGIN_DECLS
 gboolean gom_resource_get_is_from_table (GomResource *resource);
 void     gom_resource_set_is_from_table (GomResource *resource, gboolean is_from_table);
 gboolean gom_resource_has_dynamic_pkey  (GType type);
+gboolean gom_resource_do_save           (GomResource  *resource,
+                                         GomAdapter   *adapter,
+                                         GError      **error);
+void     gom_resource_build_save_cmd    (GomResource  *resource,
+                                         GomAdapter   *adapter);
 G_END_DECLS
 
 #endif /* GOM_RESOURCE_PRIV_H */
